@@ -22,6 +22,7 @@ from datetime import datetime, timedelta
 from demyst.common.config import load_config
 from demyst.common.connectors import Connectors, DemystConnectorError
 from demyst.common.track import track_function
+from demyst.common.pii import PII_REGEX
 from demyst.analytics.shiny import progressbar
 from demyst.analytics.shiny import generatehandlerbutton
 from demyst.analytics.report import report
@@ -538,7 +539,7 @@ class Analytics(object):
                                    shape="table", prefix=self.__prefix)
 
     def __field_names_from_stats():
-        categorical_stats["full_field_name"] = categorical_stats["product"].map(str) + "." + categorical_stats["flattened_name"]
+        categorical_stats["full_field_name"] = categorical_stats["product_name"].map(str) + "." + categorical_stats["flattened_name"]
         flattened_field_names = list(set(categorical_stats["full_field_name"].values))
         flattened_field_names
 
@@ -613,7 +614,6 @@ class Analytics(object):
                               { "provider_names": provider_names,
                                 "all_products": all_products })
 
-
     def __do_product_stats(self, provider_names, all_products):
         res = []
         if all_products:
@@ -632,22 +632,28 @@ class Analytics(object):
         if "id" in df_res:
             df_res.drop(columns=['id'], inplace=True)
         if "provider_id" in df_res:
-            df_res.rename(columns = {'provider_id':'product'}, inplace = True)
-        if (("flattened_name" in df_res) and ("product" in df_res)):
-            df_res["full_field_name"] = df_res["product"].map(str) + "." + df_res["flattened_name"]
-        df_res.rename(columns={'consistency_rate':'attribute_consistency_rate','entity_name':'input_entity','error_rate':'product_error_rate','field_is_populated_rate':'attribute_fill_rate','flattened_name':'attribute_flattened_name','generic_flattened_name':'attribute_generic_name','hit_rate':'product_match_rate','num_distinct_values':'attribute_unique_values','full_field_name':'attribute_full_name', 'last_updated_at': 'stats_updated_on', 'max_val': 'max_value', 'min_val': 'min_value', 'standard_deviation': 'std'}, inplace = True)
-        self.__add_extra_stats_columns(df_res)
+            df_res.rename(columns = {'provider_id':'product_name'}, inplace = True)
+        if (("flattened_name" in df_res) and ("product_name" in df_res)):
+            df_res["full_field_name"] = df_res["product_name"].map(str) + "." + df_res["flattened_name"]
+        df_res.rename(columns={'consistency_rate':'attribute_consistency_rate','entity_name':'input_entity','error_rate':'product_error_rate','field_is_populated_rate':'attribute_fill_rate','flattened_name':'attribute_name','generic_flattened_name':'attribute_generic_name','hit_rate':'product_match_rate','num_distinct_values':'unique_values','full_field_name':'attribute_full_name', 'last_updated_at': 'stats_updated_on', 'max_val': 'max_value', 'min_val': 'min_value', 'standard_deviation': 'std'}, inplace = True)
+        df_res = self.__add_extra_stats_columns(df_res, provider_names)
+        # Round to 2 places and scale to percentage
+        rnd = lambda x: round(x, 2) * 100.0
+        df_res["product_error_rate"] = df_res["product_error_rate"].apply(rnd)
+        df_res["product_match_rate"] = df_res["product_match_rate"].apply(rnd)
+        df_res["attribute_fill_rate"] = df_res["attribute_fill_rate"].apply(rnd)
         # Reorder columns
         return(df_res[[
                     'input_entity',
                     'stats_updated_on',
-                    'product',
+                    'product_name',
                     'product_error_rate',
                     'product_match_rate',
-                    'attribute_flattened_name',
+                    'attribute_name',
+                    'attribute_type',
                     'attribute_fill_rate',
                     'attribute_consistency_rate',
-                    'attribute_unique_values',
+                    'unique_values',
                     'most_common_values',
                     'std',
                     'median',
@@ -661,27 +667,36 @@ class Analytics(object):
                     'attribute_use_case'
                     ]])
 
-    def __add_extra_stats_columns(self, df):
+    def __add_extra_stats_columns(self, df, provider_names):
         public_providers_list = self.__config.auth_get(self.__config.get("MANTA_URL") + "public/providers").json()
         public_providers = {}
         for pp in public_providers_list:
             public_providers[pp["name"]] = pp
         def extract_attribute_onboarded_date(row):
-            return public_providers[row.at["product"]]["created_at"]
+            return public_providers[row.at["product_name"]]["created_at"]
         def extract_attribute_audited_date(row):
             return row.at["stats_updated_on"]
-        pii_regex = re.compile(r'(?i)(city|state|address|zip|post_code|postcode|location|locality|list_date|latitude|longitude|delivery_point|carrier_route|name|first_name|full_name|last_name|middle_name|suffix|phone|email|ssn|price|date_of_birth|date_of_death|contact|sms|court_case|sex_offender|deceased|age|owner|relation|gender|infoprivacy|tax|legal|market_value|assessed_value|business_name|businessname|companyname)')
         def extract_attribute_pii(row):
-            if pii_regex.search(row.at["attribute_flattened_name"]):
+            if PII_REGEX.search(row.at["attribute_name"]):
                 return 1
             else:
                 return 0
         def extract_attribute_use_case(row):
-            return ", ".join(map(lambda tag: tag["name"], self.__config.lookup_provider(row.at["product"])["tags"]))
+            return ", ".join(map(lambda tag: tag["name"], self.__config.lookup_provider(row.at["product_name"])["tags"]))
         df["attribute_onboarded_date"] = df.apply(extract_attribute_onboarded_date, axis=1)
         df["attribute_audited_date"] = df.apply(extract_attribute_audited_date, axis=1)
         df["attribute_pii"] = df.apply(extract_attribute_pii, axis=1)
         df["attribute_use_case"] = df.apply(extract_attribute_use_case, axis=1)
+        # Add temp column for merging, where [0] etc gets replaced with [],
+        # since that's the format used by product_outputs().
+        outputs = self.product_outputs(provider_names)
+        df["__attribute_name_no_nesting"] = df["attribute_name"].replace(r"\[(\w+)\]", "[]", regex=True)
+        df = pd.merge(df, outputs[["provider_name", "attribute", "type"]],
+                      left_on=["product_name", "__attribute_name_no_nesting"],
+                      right_on=["provider_name", "attribute"])
+        df.rename(columns={"type": "attribute_type"}, inplace=True)
+        df.drop(columns=["__attribute_name_no_nesting"], inplace=True)
+        return df
 
     # In stg, we're missing Hogfish results.  Default columns.
     def __fixup_hogfish_df(self, df):
@@ -795,7 +810,6 @@ class Analytics(object):
         """
         if hosted_input:
             inputs = self.input_file(hosted_input)
-        self.__check_provider_names(providers)
         return track_function(lambda: self.__do_enrich(providers,
                                                      inputs,
                                                      validate=validate,
@@ -819,6 +833,10 @@ class Analytics(object):
         return self.__config.auth_get(self.__config.get("MANTA_URL") + path).json()["total_cost"]
 
     def __do_enrich(self, providers, inputs, validate=True, all_updates=False, fixed_list_size=None):
+        limit = 20
+        if len(providers) > limit:
+            raise RuntimeError("Only " + str(limit) + " providers are allowed.")
+        self.__check_provider_names(providers)
         print("Starting enrichment...")
         if len(inputs) == 0:
             raise RuntimeError("Inputs dataframe is empty")
@@ -1628,10 +1646,11 @@ class Analytics(object):
                                  "temp_path": temp_path })
 
             def fetch_chunk(export):
-                req = requests.get(export["export_link"])
+                req = requests.get(export["export_link"], stream=True)
                 if req.status_code == 200:
                     with open(export["temp_path"], "wb") as f:
-                        f.write(req.content)
+                        for chunk in req:
+                            f.write(chunk)
                 else:
                     raise RuntimeError("Download of file " + export["export_link"] + " failed: " + str(req.status_code))
             try:
